from tinynn.core import initializer
from tinynn.core import layer
from tinynn.core import loss
from tinynn.core import model
from tinynn.core import net
from tinynn.core import optimizer

from tinynn.utils import data_iterator
from tinynn.utils import dataset
from tinynn.utils import downloader
from tinynn.utils import math
from tinynn.utils import metric
from tinynn.utils import seeder
from tinynn.utils import structured_param



"""Numpy random seeder"""

import numpy as np


def random_seed(seed):
    seed = int(seed)
    if seed < 0 or seed > (2 ** 32 - 1):
        raise ValueError("Seed must be between 0 and 2**32 - 1")
    np.random.seed(seed)



"""Methods to compute common machine learning metrics"""

import numpy as np


def _roc_curve(preds, targets, partition, pos_class, neg_class):
    """ROC curve (for binary classification only)"""
    thresholds = np.arange(0.0, 1.0, 1.0 / partition)[::-1]
    fpr_list, tpr_list = [], []
    for threshold in thresholds:
        pred_class = (preds >= threshold).astype(int)
        false_pos = np.sum((pred_class == pos_class) & (targets == neg_class))
        fpr_list.append(false_pos / np.sum(targets == neg_class))
        true_pos = np.sum((pred_class == pos_class) & (targets == pos_class))
        tpr_list.append(true_pos / np.sum(targets == pos_class))
    return fpr_list, tpr_list, thresholds


def auc_roc_curve(preds, targets, partition=300, pos_class=1, neg_class=0):
    """Area unser the ROC curve (for binary classification only)"""
    fprs, tprs, thresholds = _roc_curve(preds, targets, partition,
                                        pos_class, neg_class)
    auc_ = 0.0
    for i in range(len(thresholds) - 1):
        auc_ += tprs[i] * (fprs[i + 1] - fprs[i])
    info = {"fprs": fprs, "tprs": tprs, "thresholds": thresholds}
    return auc_, info


def auc(preds, targets, pos_class=1, neg_class=0):
    num_pos = np.sum(targets == pos_class)
    num_neg = np.sum(targets == neg_class)
    idx = np.argsort(preds, axis=0)
    sorted_targets = targets[idx]
    cnt = 0
    for i, target in enumerate(sorted_targets):
        if target == pos_class:
            cnt += np.sum(sorted_targets[:i] == neg_class)
    auc_ = 1.0 * cnt / (num_pos * num_neg)
    return auc_, {}


def accuracy(preds, targets):
    total_num = len(preds)
    hit_num = int(np.sum(preds == targets))
    accuracy = 1.0 * hit_num / total_num
    info = {"hit_num": hit_num, "total_num": total_num}
    return accuracy, info


def log_loss(preds, targets):
    assert len(preds) == len(targets)
    preds = np.asarray(preds)
    targets = np.asarray(targets)
    log_loss_ = np.mean(-targets * np.log(preds) -
                        (1 - targets) * np.log(1 - preds))
    return log_loss_, {}


def precision(preds, targets, pos_class=1, neg_class=0):
    """precision = TP / (TP + FP)"""
    assert len(preds) == len(targets)
    true_pos = np.sum((preds == pos_class) & (targets == pos_class))
    false_pos = np.sum((preds == pos_class) & (targets == neg_class))
    precision_ = 1. * true_pos / (true_pos + false_pos)
    info = {"true_positive": true_pos, "false_positive": false_pos}
    return precision_, info


def recall(preds, targets, pos_class=1, neg_class=0):
    """recall = TP / (TP + FN)"""
    assert len(preds) == len(targets)
    true_pos = np.sum((preds == pos_class) & (targets == pos_class))
    false_neg = np.sum((preds == neg_class) & (targets == pos_class))
    recall_ = 1. * true_pos / (true_pos + false_neg)
    info = {"true_positive": true_pos, "false_negative": false_neg}
    return recall_, info


def f1_score(preds, targets, pos_class=1, neg_class=0):
    precision_, _ = precision(preds, targets, pos_class, neg_class)
    recall_, _ = recall(preds, targets, pos_class, neg_class)
    f1_ = 2 * (precision_ * recall_) / (precision_ + recall_)
    info = {"precision": precision_, "recall": recall_}
    return f1_, info


def explained_variation(preds, targets):
    """
    Computes fraction of variance that pred_y explains about y.
    Returns 1 - Var[y-pred_y] / Var[y]

    Interpretation:
        EV=0  =>  might as well have predicted zero
        EV=1  =>  perfect prediction
        EV<0  =>  worse than just predicting zero
    """
    assert preds.shape == targets.shape
    if preds.ndim == 1:
        diff_var = np.var(targets - preds)
        target_var = np.var(targets)
        ev_ = 1.0 - diff_var / target_var
    elif preds.ndim == 2:
        diff_var = np.var(targets - preds, axis=0)
        target_var = np.var(targets, axis=0)
        non_zero_idx = np.where(target_var != 0)[0]
        ev_ = np.mean(1.0 - diff_var[non_zero_idx] / target_var[non_zero_idx])
    return ev_, {}


def r_square(preds, targets):
    assert preds.shape == targets.shape
    ss_residual = np.sum((targets - preds) ** 2, axis=0)
    ss_total = np.sum((targets - np.mean(targets, axis=0)) ** 2, axis=0)
    r_square_ = np.mean(1 - ss_residual / ss_total)
    return r_square_, {}


def mean_square_error(preds, targets):
    assert preds.shape == targets.shape
    if preds.ndim == 1:
        mse_ = np.mean(np.square(preds - targets))
    elif preds.ndim == 2:
        mse_ = np.mean(np.sum(np.square(preds - targets), axis=1))
    else:
        raise ValueError("preds supposes to have 1 or 2 dim.")
    return mse_, {}


def mean_absolute_error(preds, targets):
    assert preds.shape == targets.shape
    if preds.ndim == 1:
        mae_ = np.mean(np.abs(preds - targets))
    elif preds.ndim == 2:
        mae_ = np.mean(np.sum(np.abs(preds - targets), axis=1))
    else:
        raise ValueError("preds supposes to have 1 or 2 dim.")
    return mae_, {}



import copy

import numpy as np


class StructuredParam:
    """A helper class represents network parameters or gradients."""

    def __init__(self, param_list, nt_param_list=None):
        self.param_list = param_list
        self.nt_param_list = nt_param_list

    @property
    def values(self):
        return np.array([v for p in self.param_list for v in p.values()],
                        dtype=object)

    @values.setter
    def values(self, values):
        i = 0
        for param in self.param_list:
            for name in param.keys():
                param[name] = values[i]
                i += 1

    @property
    def nt_values(self):
        return np.array([v for p in self.nt_param_list for v in p.values()])

    @nt_values.setter
    def nt_values(self, values):
        i = 0
        for param in self.nt_param_list:
            for name in param.keys():
                param[name] = values[i]
                i += 1

    @property
    def shape(self):
        shape = list()
        for param in self.param_list:
            l_shape = dict()
            for key, val in param.items():
                l_shape[key] = val.shape
            shape.append(l_shape)
        shape = tuple(shape)
        return shape

    @staticmethod
    def _ensure_values(obj):
        if isinstance(obj, StructuredParam):
            obj = obj.values
        return obj

    def clip(self, min_=None, max_=None):
        obj = copy.deepcopy(self)
        obj.values = [v.clip(min_, max_) for v in self.values]
        return obj

    def __add__(self, other):
        obj = copy.deepcopy(self)
        obj.values = self.values + self._ensure_values(other)
        return obj

    def __radd__(self, other):
        obj = copy.deepcopy(self)
        obj.values = self._ensure_values(other) + self.values
        return obj

    def __iadd__(self, other):
        self.values += self._ensure_values(other)
        return self

    def __sub__(self, other):
        obj = copy.deepcopy(self)
        obj.values = self.values - self._ensure_values(other)
        return obj

    def __rsub__(self, other):
        obj = copy.deepcopy(self)
        obj.values = self._ensure_values(other) - self.values
        return obj

    def __isub__(self, other):
        other = self._ensure_values(other)
        self.values -= self._ensure_values(other)
        return self

    def __mul__(self, other):
        obj = copy.deepcopy(self)
        obj.values = self.values * self._ensure_values(other)
        return obj

    def __rmul__(self, other):
        obj = copy.deepcopy(self)
        obj.values = self._ensure_values(other) * self.values
        return obj

    def __imul__(self, other):
        self.values *= self._ensure_values(other)
        return self

    def __truediv__(self, other):
        obj = copy.deepcopy(self)
        obj.values = self.values / self._ensure_values(other)
        return obj

    def __rtruediv__(self, other):
        obj = copy.deepcopy(self)
        obj.values = self._ensure_values(other) / self.values
        return obj

    def __itruediv__(self, other):
        self.values /= self._ensure_values(other)
        return self

    def __pow__(self, other):
        obj = copy.deepcopy(self)
        obj.values = self.values ** self._ensure_values(other)
        return obj

    def __ipow__(self, other):
        self.values **= self._ensure_values(other)
        return self

    def __neg__(self):
        obj = copy.deepcopy(self)
        obj.values = -self.values
        return obj

    def __len__(self):
        return len(self.values)

    def __lt__(self, other):
        obj = copy.deepcopy(self)
        other = self._ensure_values(other)

        if isinstance(other, (float, int)):
            obj.values = [v < other for v in self.values]
        else:
            obj.values = [v < other[i] for i, v in enumerate(self.values)]
        return obj

    def __gt__(self, other):
        obj = copy.deepcopy(self)
        other = self._ensure_values(other)

        if isinstance(other, (float, int)):
            obj.values = [v > other for v in self.values]
        else:
            obj.values = [v > other[i] for i, v in enumerate(self.values)]
        return obj

    def __le__(self, other):
        obj = copy.deepcopy(self)
        other = self._ensure_values(other)

        if isinstance(other, (float, int)):
            obj.values = [v <= other for v in self.values]
        else:
            obj.values = [v <= other[i] for i, v in enumerate(self.values)]
        return obj

    def __ge__(self, other):
        obj = copy.deepcopy(self)
        other = self._ensure_values(other)

        if isinstance(other, (float, int)):
            obj.values = [v >= other for v in self.values]
        else:
            obj.values = [v >= other[i] for i, v in enumerate(self.values)]
        return obj

    def __and__(self, other):
        obj = copy.deepcopy(self)
        obj.values = self._ensure_values(other) & self.values
        return obj

    def __or__(self, other):
        obj = copy.deepcopy(self)
        obj.values = self._ensure_values(other) | self.values
        return obj



"""Feed-forward Neural Network class."""

import copy

from tinynn.utils.structured_param import StructuredParam


class Net:

    def __init__(self, layers):
        self.layers = layers
        self._is_training = True

    def __repr__(self):
        return "\n".join([str(layer) for layer in self.layers])

    def forward(self, inputs):
        for layer in self.layers:
            inputs = layer.forward(inputs)
        return inputs

    def backward(self, grad):
        # back propagation
        for layer in reversed(self.layers):
            grad = layer.backward(grad)

        # structured gradients
        param_grads = [copy.deepcopy(layer.grads) for layer in self.layers]
        struct_grads = StructuredParam(param_grads)
        # save the gradients w.r.t the input
        struct_grads.wrt_input = grad
        return struct_grads

    @property
    def params(self):
        trainable = [layer.params for layer in self.layers]
        non_trainable = [layer.nt_params for layer in self.layers]
        return StructuredParam(trainable, non_trainable)

    @params.setter
    def params(self, params):
        self.params.values = params.values
        self.params.nt_values = params.nt_values

    @property
    def is_training(self):
        return self._is_training

    @is_training.setter
    def is_training(self, is_training):
        for layer in self.layers:
            layer.is_training = is_training
        self._is_training = is_training



"""Common datasets"""

import gzip
import os
import pickle
import struct
import tarfile

import numpy as np

from tinynn.utils.downloader import download_url


class Dataset:

    def __init__(self, data_dir, **kwargs):
        self._train_set = None
        self._valid_set = None
        self._test_set = None

        self._save_paths = [os.path.join(data_dir, url.split("/")[-1])
                            for url in self._urls]

        self._download()
        self._parse(**kwargs)  # lgtm [py/init-calls-subclass]

    def _download(self):
        for url, checksum, save_path in zip(
                self._urls, self._checksums, self._save_paths):
            download_url(url, save_path, checksum)

    def _parse(self, **kwargs):
        raise NotImplementedError

    @property
    def train_set(self):
        return self._train_set

    @property
    def valid_set(self):
        return self._valid_set

    @property
    def test_set(self):
        return self._test_set

    @staticmethod
    def get_one_hot(targets, n_classes):
        return np.eye(n_classes, dtype=np.float32)[np.array(targets).reshape(-1)]


class MNIST(Dataset):

    def __init__(self, data_dir, one_hot=True):
        self._urls = ("https://raw.githubusercontent.com/mnielsen/neural-networks-and-deep-learning/master/data/mnist.pkl.gz",)
        self._checksums = ("98100ca27dc0e07ddd9f822cf9d244db",)
        self._n_classes = 10
        super().__init__(data_dir, one_hot=one_hot)

    def _parse(self, **kwargs):
        save_path = self._save_paths[0]
        with gzip.open(save_path, "rb") as f:
            train, valid, test = pickle.load(f, encoding="latin1")

        if kwargs["one_hot"]:
            train = (train[0], self.get_one_hot(train[1], self._n_classes))
            valid = (valid[0], self.get_one_hot(valid[1], self._n_classes))
            test = (test[0], self.get_one_hot(test[1], self._n_classes))

        self._train_set, self._valid_set, self._test_set = train, valid, test


class FashionMNIST(Dataset):

    def __init__(self, data_dir, one_hot=True):
        base_url = "http://fashion-mnist.s3-website.eu-central-1.amazonaws.com/"
        self._urls = [base_url + "train-images-idx3-ubyte.gz",
                      base_url + "train-labels-idx1-ubyte.gz",
                      base_url + "t10k-images-idx3-ubyte.gz",
                      base_url + "t10k-labels-idx1-ubyte.gz"]
        self._checksums = ["8d4fb7e6c68d591d4c3dfef9ec88bf0d",
                           "25c81989df183df01b3e8a0aad5dffbe",
                           "bef4ecab320f06d8554ea6380940ec79",
                           "bb300cfdad3c16e7a12a480ee83cd310"]
        self._n_classes = 10
        super().__init__(data_dir, one_hot=one_hot)

    @staticmethod
    def read_idx(filename):
        with gzip.open(filename, "rb") as fileobj:
            _, _, dims = struct.unpack(">HBB", fileobj.read(4))
            shape = tuple(struct.unpack(">I", fileobj.read(4))[0]
                          for d in range(dims))
            return np.frombuffer(fileobj.read(), dtype=np.uint8).reshape(shape)

    def _parse(self, **kwargs):
        train_x, train_y, test_x, test_y = (self.read_idx(path)
                                            for path in self._save_paths)
        # normalize
        train_x = train_x.astype(float) / 255.0
        test_x = test_x.astype(float) / 255.0
        train_x = train_x.reshape((len(train_x), -1))
        test_x = test_x.reshape((len(test_x), -1))

        if kwargs["one_hot"]:
            train_y = self.get_one_hot(train_y, self._n_classes)
            test_y = self.get_one_hot(test_y, self._n_classes)
        self._train_set = (train_x, train_y)
        self._test_set = (test_x, test_y)


class Cifar(Dataset):

    @staticmethod
    def _cifar_normalize(data):
        means, stds = (0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)
        data = data.reshape((len(data), -1, 3))
        for ch in range(3):
            data[:, :, ch] = (data[:, :, ch] - means[ch]) / stds[ch]
        data = data.reshape((len(data), -1))
        return data

    def _parse(self, **kwargs):
        raise NotImplementedError

    def _parse_tarfile(self):
        dataset = {}
        with open(self._save_paths[0], "rb") as f:
            tar = tarfile.open(fileobj=f)
            for item in tar:
                obj = tar.extractfile(item)
                if not obj or item.size < 100:
                    continue
                cont = pickle.load(obj, encoding="bytes")
                dataset[item.name.split("/")[-1]] = cont
        return dataset


class Cifar10(Cifar):

    def __init__(self, data_dir, one_hot=False, normalize=False):
        self._urls = ("https://www.cs.toronto.edu/~kriz/cifar-10-python.tar.gz",)
        self._checksums = ("c58f30108f718f92721af3b95e74349a",)
        self._n_classes = 10
        super().__init__(data_dir, one_hot=one_hot, normalize=normalize)

    def _parse(self, **kwargs):
        dataset = self._parse_tarfile()
        data_batch_name = [f"data_batch_{i}" for i in range(1, 6)]
        train_x, train_y = [], []
        for name in data_batch_name:
            cont = dataset[name]
            train_x.append(cont[b"data"])
            train_y.extend(cont[b"labels"])
        train_x = np.concatenate(train_x, axis=0)

        train_x = train_x / 255.0
        train_y = np.asarray(train_y)
        test_x = dataset["test_batch"][b"data"] / 255.
        test_y = np.asarray(dataset["test_batch"][b"labels"])

        if kwargs["normalize"]:
            train_x = self._cifar_normalize(train_x)
            test_x = self._cifar_normalize(test_x)

        if kwargs["one_hot"]:
            train_y = self.get_one_hot(train_y, self._n_classes)
            test_y = self.get_one_hot(test_y, self._n_classes)

        self._train_set = (train_x, train_y)
        self._test_set = (test_x, test_y)


class Cifar100(Cifar):

    def __init__(self, data_dir, one_hot=False, normalize=False):
        self._urls = ("https://www.cs.toronto.edu/~kriz/cifar-100-python.tar.gz",)
        self._checksums = ("eb9058c3a382ffc7106e4002c42a8d85",)
        self._n_classes = 100
        super().__init__(data_dir, one_hot=one_hot, normalize=normalize)

    def _parse(self, **kwargs):
        dataset = self._parse_tarfile()
        train_x = dataset["train"][b"data"] / 255.0
        train_y = np.asarray(dataset["train"][b"fine_labels"])

        test_x = dataset["test"][b"data"] / 255.0
        test_y = np.asarray(dataset["test"][b"fine_labels"])

        if kwargs["normalize"]:
            train_x = self._cifar_normalize(train_x)
            test_x = self._cifar_normalize(test_x)

        if kwargs["one_hot"]:
            train_y = self.get_one_hot(train_y, self._n_classes)
            test_y = self.get_one_hot(test_y, self._n_classes)

        self._train_set = (train_x, train_y)
        self._test_set = (test_x, test_y)



"""Simple utilities to download and save a file with progress bar."""

import hashlib
import os
from urllib.error import URLError
from urllib.request import urlretrieve


def show_progress(blk_num, blk_sz, tot_sz):
    percentage = 100.0 * blk_num * blk_sz / tot_sz
    print(f"Progress: {percentage:.1f} %", end="\r", flush=True)


def md5_checksum(file_path):
    with open(file_path, "rb") as fileobj:
        checksum = hashlib.md5(fileobj.read()).hexdigest()
    return checksum


def download_url(url, file_path, checksum):
    file_dir = os.path.dirname(file_path)
    if not os.path.exists(file_dir):
        os.makedirs(file_dir)

    if os.path.exists(file_path):
        if md5_checksum(file_path) == checksum:
            print(f"{file_path} already exists.")
            return
        print("Wrong checksum!")

    try:
        print(f"Downloading {url} to {file_path}")
        urlretrieve(url, file_path, show_progress)
    except URLError:
        raise RuntimeError("Error downloading resource!")
    except KeyboardInterrupt:
        print("Interrupted")



"""Various of network parameter initializers."""

import numpy as np


def get_fans(shape):
    fan_in = shape[0] if len(shape) == 2 else np.prod(shape[1:])
    fan_out = shape[1] if len(shape) == 2 else shape[0]
    return fan_in, fan_out


class Initializer:

    def __call__(self, shape):
        return self.init(shape).astype(np.float32)

    def init(self, shape):
        raise NotImplementedError


class Normal(Initializer):

    def __init__(self, mean=0.0, std=1.0):
        self._mean = mean
        self._std = std

    def init(self, shape):
        return np.random.normal(loc=self._mean, scale=self._std, size=shape)


class TruncatedNormal(Initializer):

    def __init__(self, low, high, mean=0.0, std=1.0):
        self._mean, self._std = mean, std
        self._low, self._high = low, high

    def init(self, shape):
        data = np.random.normal(loc=self._mean, scale=self._std, size=shape)
        while True:
            mask = (data > self._low) & (data < self._high)
            if mask.all():
                break
            data[~mask] = np.random.normal(loc=self._mean, scale=self._std,
                                           size=(~mask).sum())
        return data


class Uniform(Initializer):

    def __init__(self, a=0.0, b=1.0):
        self._a = a
        self._b = b

    def init(self, shape):
        return np.random.uniform(low=self._a, high=self._b, size=shape)


class Constant(Initializer):

    def __init__(self, val):
        self._val = val

    def init(self, shape):
        return np.full(shape=shape, fill_value=self._val)


class Zeros(Constant):

    def __init__(self):
        super(Zeros, self).__init__(0.0)


class Ones(Constant):

    def __init__(self):
        super(Ones, self).__init__(1.0)


class XavierUniform(Initializer):
    """
    Implement the Xavier method described in
    "Understanding the difficulty of training deep feedforward neural networks"
    Glorot, X. & Bengio, Y. (2010)

    Weights will have values sampled from uniform distribution U(-a, a) where
    a = gain * sqrt(6.0 / (num_in + num_out))

    """

    def __init__(self, gain=1.0):
        self._gain = gain

    def init(self, shape):
        fan_in, fan_out = get_fans(shape)
        a = self._gain * np.sqrt(6.0 / (fan_in + fan_out))
        return np.random.uniform(low=-a, high=a, size=shape)


class XavierNormal(Initializer):
    """
    Implement the Xavier method described in
    "Understanding the difficulty of training deep feedforward neural networks"
    Glorot, X. & Bengio, Y. (2010)

    Weights will have values sampled from uniform distribution N(0, std) where
    std = gain * sqrt(1.0 / (num_in + num_out))
    """

    def __init__(self, gain=1.0):
        self._gain = gain

    def init(self, shape):
        fan_in, fan_out = get_fans(shape)
        std = self._gain * np.sqrt(2.0 / (fan_in + fan_out))
        return np.random.normal(loc=0.0, scale=std, size=shape)


class HeUniform(Initializer):
    """
    Implement the He initialization method described in
    "Delving deep into rectifiers: Surpassing human-level performance
    on ImageNet classification" He, K. et al. (2015)

    Weights will have values sampled from uniform distribution U(-a, a) where
    a = sqrt(6.0 / num_in)
    """

    def __init__(self, gain=1.0):
        self._gain = gain

    def init(self, shape):
        fan_in, _ = get_fans(shape)
        a = self._gain * np.sqrt(6.0 / fan_in)
        return np.random.uniform(low=-a, high=a, size=shape)


class HeNormal(Initializer):
    """
    Implement the He initialization method described in
    "Delving deep into rectifiers: Surpassing human-level performance
    on ImageNet classification" He, K. et al. (2015)

    Weights will have values sampled from normal distribution N(0, std) where
    std = sqrt(2.0 / num_in)
    """

    def __init__(self, gain=1.0):
        self._gain = gain

    def init(self, shape):
        fan_in, _ = get_fans(shape)
        std = self._gain * np.sqrt(2.0 / fan_in)
        return np.random.normal(loc=0.0, scale=std, size=shape)



"""Useful math utilities."""

import numpy as np


def softmax(x, t=1.0, axis=-1):
    x_ = x / t
    x_max = np.max(x_, axis=axis, keepdims=True)
    exps = np.exp(x_ - x_max)
    return exps / np.sum(exps, axis=axis, keepdims=True)


def log_softmax(x, t=1.0, axis=-1):
    x_ = x / t
    x_max = np.max(x_, axis=axis, keepdims=True)
    exps = np.exp(x_ - x_max)
    exp_sum = np.sum(exps, axis=axis, keepdims=True)
    return x_ - x_max - np.log(exp_sum)


def sigmoid(x):
    return 1.0 / (1.0 + np.exp(-x))



"""Loss functions"""

import numpy as np
from tinynn.utils.math import log_softmax
from tinynn.utils.math import sigmoid
from tinynn.utils.math import softmax


class Loss:

    def loss(self, *args, **kwargs):
        raise NotImplementedError

    def grad(self, *args, **kwargs):
        raise NotImplementedError


class MSE(Loss):

    def loss(self, predictions, targets):
        return 0.5 * np.sum((predictions - targets) ** 2) / targets.shape[0]

    def grad(self, predictions, targets):
        return (predictions - targets) / targets.shape[0]


class MAE(Loss):

    def loss(self, predictions, targets):
        return np.sum(np.abs(predictions - targets)) / targets.shape[0]

    def grad(self, predictions, targets):
        return np.sign(predictions - targets) / targets.shape[0]


class Huber(Loss):

    def __init__(self, delta=1.0):
        self._delta = delta

    def loss(self, predictions, targets):
        l1_dist = np.abs(predictions - targets)
        mse_mask = l1_dist < self._delta  # MSE part
        mae_mask = ~mse_mask  # MAE part
        mse = 0.5 * (predictions - targets) ** 2
        mae = self._delta * l1_dist - 0.5 * self._delta ** 2
        return np.sum(mse * mse_mask + mae * mae_mask) / targets.shape[0]

    def grad(self, predictions, targets):
        err = predictions - targets
        mse_mask = np.abs(err) < self._delta  # MSE part
        mae_mask = ~mse_mask  # MAE part
        mse_grad = err
        mae_grad = np.sign(err) * self._delta
        return (mae_grad * mae_mask + mse_grad * mse_mask) / targets.shape[0]


class SoftmaxCrossEntropy(Loss):

    def __init__(self, T=1.0, weights=None):
        self._weights = np.asarray(weights) if weights is not None else weights
        self._T = T

    def loss(self, logits, labels):
        nll = -(log_softmax(logits, t=self._T, axis=1) * labels).sum(axis=1)
        if self._weights is not None:
            nll *= self._weights[np.argmax(labels, axis=1)]
        return np.sum(nll) / labels.shape[0]

    def grad(self, logits, labels):
        grads = softmax(logits, t=self._T) - labels
        if self._weights is not None:
            grads *= self._weights
        return grads / labels.shape[0]


class SigmoidCrossEntropy(Loss):
    """let logits = a, label = y, weights[neg] = w1, weights[pos] = w2
    L = - w2 * y * log(1 / (1 + exp(-a)) - w1 * (1-y) * log(exp(-a) / (1 + exp(-a))
      = w1 * a * (1 - y) - (w2 * y - w1 * (y - 1)) * log(sigmoid(a))
    if w1 == w2 == 1:
    L = a * (1 - y) - log(sigmoid(a))

    G = w1 * sigmoid(a) - w2 * y + (w2 - w1) * y * sigmoid(a)
    if w1 == w2 == 1:
    G = sigmoid(a) - y
    """
    def __init__(self, weights=None):
        weights = np.ones(2, dtype=np.float32) if weights is None else weights
        self._weights = np.asarray(weights)

    def loss(self, logits, labels):
        neg_weight, pos_weight = self._weights
        cost = neg_weight * logits * (1 - labels) - \
               (pos_weight * labels - neg_weight * (labels - 1)) * \
               np.log(sigmoid(logits))
        return np.sum(cost) / labels.shape[0]

    def grad(self, logits, labels):
        neg_weight, pos_weight = self._weights
        grads = neg_weight * sigmoid(logits) - pos_weight * labels + \
                (pos_weight - neg_weight) * labels * sigmoid(logits)
        return grads / labels.shape[0]



"""Various optimization algorithms and learning rate schedulers."""

import numpy as np


class Optimizer:

    def __init__(self, lr, weight_decay):
        self.lr = lr
        self.weight_decay = weight_decay

    def step(self, grads, params):
        # compute the gradient step
        grads = self.compute_step(grads)
        # apply weight_decay if specified
        if self.weight_decay:
            grads -= self.lr * self.weight_decay * params
        # take a step
        params += grads

    def compute_step(self, grads):
        grads.values = self._compute_step(grads.values)
        return grads

    def _compute_step(self, grads):
        raise NotImplementedError


class SGD(Optimizer):

    def __init__(self, lr=0.01, weight_decay=0.0):
        super().__init__(lr, weight_decay)

    def _compute_step(self, grads):
        return -self.lr * grads


class Adam(Optimizer):

    def __init__(self,
                 lr=0.001,
                 beta1=0.9,
                 beta2=0.999,
                 epsilon=1e-8,
                 weight_decay=0.0):
        super().__init__(lr, weight_decay)
        self._b1 = beta1
        self._b2 = beta2
        self._epsilon = epsilon

        self._t = 0
        self._m = 0
        self._v = 0

    def _compute_step(self, grads):
        self._t += 1

        self._m += (1.0 - self._b1) * (grads - self._m)
        self._v += (1.0 - self._b2) * (grads ** 2 - self._v)

        # bias correction
        _m = self._m / (1 - self._b1 ** self._t)
        _v = self._v / (1 - self._b2 ** self._t)

        step = -self.lr * _m / (_v ** 0.5 + self._epsilon)
        return step


class RAdam(Optimizer):
    """Rectified Adam. Ref: https://arxiv.org/pdf/1908.03265v1.pdf """
    def __init__(self,
                 lr=0.001,
                 beta1=0.9,
                 beta2=0.999,
                 epsilon=1e-8,
                 weight_decay=0.0):
        super().__init__(lr, weight_decay)
        self._b1 = beta1
        self._b2 = beta2
        self._epsilon = epsilon

        self._t = 0
        self._m = 0
        self._v = 0

        self.rho = 2.0 / (1 - self._b2) - 1.0

    def _compute_step(self, grads):
        self._t += 1

        self._m += (1.0 - self._b1) * (grads - self._m)
        self._v += (1.0 - self._b2) * (grads ** 2 - self._v)

        # bias correction
        _m = self._m / (1 - self._b1 ** self._t)

        _rho = self.rho - 2 * self._b2 ** self._t / (1 - self._b2 ** self._t)
        if _rho > 4.0:
            _v = self._v / (1 - self._b2 ** self._t)
            _r = (((_rho - 4) * (_rho - 2) * self.rho) / \
                    ((self.rho - 4) * (self.rho - 2) * _rho)) ** 0.5
            step = -self.lr * _m * _r / (_v ** 0.5 + self._epsilon)
        else:
            step = -self.lr * _m
        return step


class RMSProp(Optimizer):
    """Root Mean Square Prop optimizer
    mean_square = decay * mean_square{t-1} + (1-decay) * grad_t**2
    mom = momentum * mom{t-1} + lr * grad_t / sqrt(mean_square + epsilon)
    """
    def __init__(self,
                 lr=0.01,
                 decay=0.99,
                 momentum=0.0,
                 epsilon=1e-8,
                 weight_decay=0.0):
        super().__init__(lr, weight_decay)
        self._rho = decay
        self._momentum = momentum
        self._epsilon = epsilon

        self._rms = 0
        self._mom = 0

    def _compute_step(self, grads):
        self._rms += (1 - self._rho) * (grads ** 2 - self._rms)
        self._mom = self._momentum * self._mom + self.lr * grads / \
                (self._rms + self._epsilon) ** 0.5
        step = -self._mom
        return step


class Momentum(Optimizer):
    """accumulation = momentum * accumulation + gradient
    variable -= learning_rate * accumulation
    """
    def __init__(self, lr, momentum=0.9, weight_decay=0.0):
        super().__init__(lr, weight_decay)
        self._momentum = momentum
        self._acc = 0

    def _compute_step(self, grads):
        self._acc = self._momentum * self._acc + grads
        step = -self.lr * self._acc
        return step


class Adagrad(Optimizer):
    """AdaGrad optimizer
    accumulation = - (learning_rate / sqrt(G + epsilon)) * gradient
    where G is the element-wise sum of square gradient
    ref: http://www.jmlr.org/papers/volume12/duchi11a/duchi11a.pdf
    """
    def __init__(self, lr, epsilon=1e-8, weight_decay=0.0):
        super().__init__(lr, weight_decay)
        self._g = 0
        self._epsilon = epsilon

    def _compute_step(self, grads):
        self._g += grads ** 2
        adjust_lr = self.lr / (self._g + self._epsilon) ** 0.5
        step = -adjust_lr * grads
        return step


class Adadelta(Optimizer):
    """Adadelta algorithm (https://arxiv.org/abs/1212.5701)"""
    def __init__(self, lr=1.0, decay=0.9, epsilon=1e-8, weight_decay=0.0,):
        super().__init__(lr, weight_decay)
        self._epsilon = epsilon
        self._rho = decay
        self._rms = 0  # running average of square gradient
        self._delta = 0  # running average of delta

    def _compute_step(self, grads):
        self._rms += (1 - self._rho) * (grads ** 2 - self._rms)
        std = (self._delta + self._epsilon) ** 0.5
        delta = grads * (std / (self._rms + self._epsilon) ** 0.5)
        step = - self.lr * delta
        self._delta += (1 - self._rho) * (delta ** 2 - self._delta)
        return step


class BaseScheduler:
    """BaseScheduler model receive a optimizer and Adjust the lr
    by calling step() method during training.
    """
    def __init__(self, optimizer):
        self._optimizer = optimizer
        self._init_lr = self.curr_lr

        self._t = 0

    def step(self):
        self._t += 1
        self._optimizer.lr = self._compute_lr()
        return self.curr_lr

    def _compute_lr(self):
        raise NotImplementedError

    @property
    def curr_lr(self):
        return self._optimizer.lr


class StepLR(BaseScheduler):
    """LR decayed by gamma every "step_size" epochs."""
    def __init__(self,
                 optimizer,
                 step_size,
                 gamma=0.1):
        super().__init__(optimizer)
        assert step_size >= 1
        self._step_size = step_size
        self._gamma = gamma

    def _compute_lr(self):
        decay = self._gamma if self._t % self._step_size == 0 else 1.0
        return decay * self.curr_lr


class MultiStepLR(BaseScheduler):
    """LR decayed by gamma when #steps reaches one of the milestones.
    Milestones must be monotonically increasing.
    """
    def __init__(self, optimizer, milestones, gamma=0.1):
        super().__init__(optimizer)
        milestones = [int(m) for m in milestones]
        assert len(milestones) > 0, "milestones requires at-least one element!"
        assert all(x < y for x, y in zip(milestones[:-1], milestones[1:])), \
               "milestones must be a list of int and be increasing!"

        self._milestones = milestones
        self._gamma = gamma

    def _compute_lr(self):
        decay = self._gamma if self._t in self._milestones else 1.0
        return decay * self.curr_lr


class ExponentialLR(BaseScheduler):
    """ExponentialLR is computed by:
    lr_decayed = lr * decay_rate ^ (current_steps / decay_steps)
    """
    def __init__(self,
                 optimizer,
                 decay_steps,
                 decay_rate=(1. / np.e)):
        super().__init__(optimizer)
        self._decay_steps = decay_steps
        self._decay_rate = decay_rate

    def _compute_lr(self):
        if self._t <= self._decay_steps:
            decay = self._decay_rate ** (self._t / self._decay_steps)
            return self._init_lr * decay
        return self.curr_lr


class LinearLR(BaseScheduler):
    """Linear decay learning rate when the number of the epoch is in
    [start_step, start_step + decay_steps]
    """
    def __init__(self,
                 optimizer,
                 decay_steps,
                 final_lr=1e-6,
                 start_step=0):
        super().__init__(optimizer)
        assert decay_steps > 0

        self._lr_delta = (final_lr - self._init_lr) / decay_steps

        self._final_lr = final_lr
        self._decay_steps = decay_steps
        self._start_step = start_step

    def _compute_lr(self):
        if self._t > self._start_step:
            if self._t <= self._start_step + self._decay_steps:
                return self.curr_lr + self._lr_delta
        return self.curr_lr


class CyclicalLR(BaseScheduler):
    """Cyclical increase and decrease learning rate within a reasonable range.
    Ref: https://arxiv.org/pdf/1506.01186.pdf
    """
    def __init__(self,
                 optimizer,
                 cyclical_steps,
                 min_lr=1e-3,
                 max_lr=1e-2):
        super().__init__(optimizer)
        assert cyclical_steps > 2
        assert max_lr >= min_lr
        self._cyclical_steps = cyclical_steps
        self._min_lr = min_lr
        self._max_lr = max_lr
        self._abs_lr_delta = 2 * (max_lr - min_lr) / cyclical_steps

        self._is_cycling = False
        self._cycling_start_t = None

    def _compute_lr(self):
        if self.curr_lr > self._max_lr:
            return self.curr_lr - self._abs_lr_delta
        if self.curr_lr < self._min_lr:
            return self.curr_lr + self._abs_lr_delta

        if not self._is_cycling:
            self._is_cycling = True
            self._cycling_start_t = self._t

        if ((self._t - self._cycling_start_t) % self._cyclical_steps <
                self._cyclical_steps // 2):
            return self.curr_lr + self._abs_lr_delta
        return self.curr_lr - self._abs_lr_delta
